#include "retrohack.h"
#include <math.h>

#ifndef M_PI
#define M_PI 3.141592654f
#endif

/*
	Pixels is an array of 1x1 graphics, in the 16 different CbCr combos.

	CreatePixels creates the graphics, FreePixels draws them
*/
void *Pixels[16];

void CreatePixels(void)
{
	int c = 16;
	unsigned char Data;
	while(c--)
	{
		Data = c;
		Pixels[c] = rt_UploadGraphic(RT_Y, &Data, 1, 1, 1, c^1);
	}
}

void FreePixels(void)
{
	int c = 16;
	while(c--)
		rt_FreeGraphic(Pixels[c]);
}

/*
	DrawTriangle draws a triangle between V1, V2 and V3 (all fixed point) by decomposing it into
	two sprites. This is exactly how Steel Talons, Hard Drivin', etc managed it on the Lynx
*/
void DrawTriangle(int *V1, int *V2, int *V3, int colour)
{
	int CurWidth, CurX;
	int XAdd;

	// sort V1 to V3 from top to bottom
	if(V1[1] > V2[1])
	{
		int *T = V2;
		V2 = V1;
		V1 = T;
	}
	if(V2[1] > V3[1])
	{
		int *T = V2;
		V2 = V3;
		V3 = T;
	}
	if(V1[1] > V2[1])
	{
		int *T = V2;
		V2 = V1;
		V1 = T;
	}

	// do some vertex snapping to alleviate precision problems later on
	V1[0] &= ~63; V1[1] &= ~63;
	V2[0] &= ~63; V2[1] &= ~63;
	V3[0] &= ~63; V3[1] &= ~63;

	// if this polygon has zero height, return
	if(V3[1] == V1[1]) return;

	XAdd = fdiv(V3[0] - V1[0], V3[1] - V1[1]);

	// do a top part, if needed
	if(V1[1] != V2[1])
	{
		int ScaleAdd;

		// work out intersection on other line
		CurX = V1[0] + fixmul(XAdd, V2[1] - V1[1]);

		// hence work out ScaleAdd
		CurWidth = V2[0] - CurX;
		ScaleAdd = fixdiv(CurWidth, V2[1] - V1[1]);

		// draw graphic, get ready for second half
		rt_Blit(Pixels[colour >> 4], colour&0xf, V1[0], V1[1], 0, V2[1] - V1[1], XAdd, ScaleAdd);
	}
	else
	{
		CurWidth = V2[0] - V1[0];
		CurX = V1[0];
	}

	// do a bottom part, if needed
	if(V3[1] != V2[1])
	{
		// hence work out ScaleAdd & Xadd
		int ScaleAdd = fixdiv(-CurWidth, V3[1] - V2[1]);

		// draw graphic
		rt_Blit(Pixels[colour >> 4], colour&0xf, CurX, V2[1], CurWidth, V3[1] - V2[1], XAdd, ScaleAdd);
	}
}

/*
	ClearScreen stretches one of the pixel graphics up to fill the screen, thereby clearing it
*/
void ClearScreen(int colour)
{
	rt_Blit(Pixels[colour >> 4], colour&0xf, 0, 0, itofix(320), itofix(240), 0, 0);
}

int main(int argc, const char *argv[])
{
	// initialise machine
	rt_Init();

	// create the pixel graphics we'll use for screen clearing and triangle drawing
	CreatePixels();
	
	// set the program title
	rtSim_SetWindowTitle("RetroHack: polygons");

	// loop until the user requests a quit
	while(!rtSim_QuitWanted())
	{
		// Angle keeps track of the current rotation, Verts is used to calculate vertex positions
		static float Angle = 0.0f;
		int Verts[3][2];

		// some trivial sin/cos to make a circumscribed, equilateral triangle
		int c = 3;
		while(c--)
		{
			float A = Angle + ((float)c / 1.5f) * M_PI;
			Verts[c][0] = itofix(160) + ftofix(80*sin(A));
			Verts[c][1] = itofix(120) + ftofix(80*cos(A));
		}

		// increment angle, to achieve rotation
		Angle += 0.01f;

		// clear to black
		ClearScreen(10);

		// draw triangle
		DrawTriangle(Verts[0], Verts[1], Verts[2], 15);

		// flip buffers
		rt_Flip();
	}

	// free pixels
	FreePixels();

	// exit RetroHack simulator
	rt_Exit();

	return 0;
}
END_OF_MAIN()
