#include "retrohack.h"
#include <stdio.h>

// Volumes will hold a table that maps 16 bit sound levels to volume values for the three AY channels
int Volumes[65536][3];

// the AY has a logarithmic DAC, with volume levels roughly as per the following table
int Levels[16]=
{
	0, 300, 447, 634, 925, 1351, 1850, 2990, 3694, 5781, 7705, 9829, 12460, 15013, 18528, 21845
};

// this function fills the 'Volumes' table by brute force
void SetupVolumeTables(void)
{
	// subs keeps track of which entries in 'Volume' have been mapped exactly - start by assuming that none have
	int Subs[65536];
	int c = 65536;
	while(c--)
		Subs[c] = TRUE;

	// run through all the AY volume levels, works out what they map to on a linear scale, stores the combinations
	c = 4096;
	while(c--)
	{
		int Volume = Levels[c&15] + Levels[(c >> 4)&15] + Levels[c >> 8];
		if(Subs[Volume])
		{
			Volumes[Volume][0] = c&15;
			Volumes[Volume][1] = (c >> 4)&15;
			Volumes[Volume][2] = c >> 8;
			Subs[Volume] = FALSE;
		}
	}

	// work out which combinations aren't mapped, find the closest substitute
	c = 65536;
	while(c--)
	{
		if(Subs[c])
		{
			int upc = c, downc = c;
			// work out if it's better to round up or down
			while(Subs[upc] && Subs[downc])
			{
				upc++;
				downc--;
			}
			if(!Subs[upc])
			{
				Volumes[c][0] = Volumes[upc][0];
				Volumes[c][1] = Volumes[upc][1];
				Volumes[c][2] = Volumes[upc][2];
			}
			else
			{
				Volumes[c][0] = Volumes[downc][0];
				Volumes[c][1] = Volumes[downc][1];
				Volumes[c][2] = Volumes[downc][2];
			}
		}
	}
}

int main(int argc, const char *argv[])
{
	FILE *sample;
	int c;

	// open the sound sample, exit the program now if it is missing
	sample = fopen("seinfeld.snd", "rb");
	if(!sample)
		return -1;

	// build the volume tables, init the RetroHack simulator
	SetupVolumeTables();
	rt_Init();
	rtSim_SetWindowTitle("RetroHack: Sampled sound");

	// set all three AY channels to a constant level, so that volume twiddling works
	c = 15;
	while(c--)
	{
		ay_Write(AY_REGSELECT, c);
		ay_Write(AY_REGVALUE, 0);
	}
	ay_Write(AY_REGSELECT, 7);
	ay_Write(AY_REGVALUE, 0xff);

	// loop while the user hasn't requested an early quit and there is still sample data
	while(!rtSim_QuitWanted() && !feof(sample))
	{
		unsigned short Sample;
		static unsigned short LastSample = 0;

		// read a new sample. Samples in the file are
		Sample = fgetc(sample);
		Sample |= fgetc(sample) << 8;

		/*
			now output the three volume levels. This isn't the best way to use the AY to
			output PCM audio because the three volumes aren't changed simultaneously, which
			causes the background white noise you hear in this program. But it is still
			an effective RetroHack simulator test...
		*/
		if(Volumes[Sample][0] != Volumes[LastSample][0])
		{
			ay_Write(AY_REGSELECT, 8);
			ay_Write(AY_REGVALUE, Volumes[Sample][0]);
		}
		if(Volumes[Sample][1] != Volumes[LastSample][1])
		{
			ay_Write(AY_REGSELECT, 9);
			ay_Write(AY_REGVALUE, Volumes[Sample][1]);
		}
		if(Volumes[Sample][2] != Volumes[LastSample][2])
		{
			ay_Write(AY_REGSELECT, 10);
			ay_Write(AY_REGVALUE, Volumes[Sample][2]);
		}
		LastSample = Sample;

		/*
			wait for the next horizontal sync  use this as a means of timing.
			There are 311 * 50 = 15550 vertical syncs a second, so this means we're
			outputting a 15550 Hz wave
		*/
		rt_WaitEvent(RT_HSYNC);
	}

	// close the RetroHack simulator and the sound file
	rt_Exit();
	fclose(sample);

	return 0;
}
END_OF_MAIN()
